/**
 * server.js
 * Server initialization
 */
'use strict';
const express = require('express');
const bodyParser = require('body-parser');
const cors = require('cors');
const uuidV1 = require('uuid/v1');
const WorkflowService = require('./services/workflow/workflowservice');
const { logger } = require('./utils/logger.js');
const config = require('./config');

const rootUrl = '/*/hooks';
const app = express();

logger('INFO', 'Application NODE_ENV: ' + process.env.NODE_ENV);
logger('INFO', 'Hooks debug enabled: ' + config.hooks.debugEnabled);
logger('INFO', 'Workflow service debug enabled: ' + config.workflowService.debugEnabled);

// Workflow service creation
const workflowService = new WorkflowService(config.workflowService);

// Enable Cross Origin Resource Sharing 
app.use('*', cors());

// Stop node app
if (config.hooks.debugEnabled) {
  app.use(rootUrl + '/stop', (req, res) => {
    logger('INFO', 'Stop request');
    res.send('Stop request initiated.');
    process.exit();
  });
}

// Parse application/x-www-form-urlencoded
app.use(bodyParser.urlencoded({ extended: false }));

// Parse application/json
app.use(bodyParser.json());

// Catch json syntax error
app.use(function (error, req, res, next) {
  if (error instanceof SyntaxError) {
    logger('ERROR', 'Invalid JSON');
    res.status(400).json({ error: 'Invalid JSON.' });
  } else {
    next();
  }
});

// Hooks POST endpoint
app.post(rootUrl + '/:token' , (req, res) => {
  res.set('Content-Type', 'application/json');

  req.uuid = uuidV1();
  logger('INFO', 'New request ' + req.uuid);
  const token = req.params.token;
  var context = {
    hookToken: token,
    operationCount: 0,
    operationErrorCount: 0
  };
  var body;
  // application/x-www-form-urlencoded
  if (req.headers['content-type'] == 'application/x-www-form-urlencoded') {
    if (req.body.payload) {
      try {
        body = JSON.parse(req.body.payload);
      } catch (e) {
        logger('ERROR', 'Request ' + req.uuid + ' invalid JSON in payload');
        res.status(400).json({ error: 'Invalid JSON in payload.' });
        return;
      }
    }else {
      logger('ERROR', 'Request ' + req.uuid + ' the payload key is not defined');
      res.status(400).json({ error: 'The payload key is not defined.' });
    }
  }
  else if (req.headers['content-type'] == 'application/json') {
    body = req.body;
  }else {
    logger('ERROR', 'Request ' + req.uuid + ' the content type not supported');
    res.status(400).json({ error: 'The content type not supported.' });
  }
  if (!body) {
    logger('ERROR', 'Request ' + req.uuid + ' JSON body is null');
    res.status(400).json({ error: 'Invalid JSON.'});
  }
  if (!body.operation) {
    logger('ERROR', 'Request ' + req.uuid + ' missing operation');
    res.status(400).json({ error: 'An operation is required.'});
  } else if (typeof body.operation != 'string') {
    logger('ERROR', 'Request ' + req.uuid + ' invalid operation value');
    res.status(400).json({ error: 'Invalid operation value.'});
  }
  logger('INFO', 'Request ' + req.uuid + ' token:' + token + ' operation:' + body.operation);
  if (config.hooks.operations.indexOf(body.operation) < 0) {
    logger('ERROR', 'Request ' + req.uuid + ' operation not found');
    res.status(404).json({ error: 'The requested operation was not found.'});
    return;
  }

  workflowService.invoke(body.operation, context, body.args).then(
    result => res.json(result),
    error => {
      logger('ERROR', 'Request ' + req.uuid + ' ' + error);
      res.status(error.statusCode ? error.statusCode : 500 ).json({error: error.message });
    }
  );
});

// Start Express Server
app.listen(config.app.port, function () {
  logger('INFO', 'Hooks Server running on port ' + config.app.port);
});
