var chai = require('chai');
var chaiHttp = require('chai-http');
var expect = require('chai').expect;
var should = chai.should();
var utils = require('../../utils');
chai.use(chaiHttp);

var groupId = null;

describe('CreateGroup', function () {
  it('should create a group with required fields (name, directoryId)', function (done) {
    chai.request(url)
      .post('/')
      .send({
        'operation': 'createGroup',
        'args': {
          'input': {
            'name': 'Group F',
            'directoryId': 'RGlyZWN0b3J5OjM='
          }
        }
      })
      .end(function (err, res) {
        res.should.have.status(200);
        res.should.be.json;

        res.body.should.have.property('clientMutationId');
        res.body.should.have.property('groupId');

        expect(res.body.clientMutationId).to.be.null;
        expect(res.body.groupId).to.not.be.null;

        groupId = res.body.groupId;

        done();
      });
  });

  it("should return the group created with the 'Group F' name ", function (done) {
    if (groupId === null) {
      this.skip();
    }
    chai.request(url)
      .post('/')
      .send({
        'operation': 'getGroup',
        'args': {
          'id': groupId
        }
      })
      .end(function (err, res) {
        res.should.have.status(200);
        res.should.be.json;

        utils.validateGroupProperties(res.body);

        expect(res.body).to.include({ id: groupId, name: 'Group F', description: null, directoryId: 'RGlyZWN0b3J5OjM=' });

        done();
      });
  });

  it('should create a group with some pertinent fields', function (done) {
    chai.request(url)
      .post('/')
      .send({
        'operation': 'createGroup',
        'args': {
          'input': {
            'name': 'Group G',
            'description': 'Group G',
            'directoryId': 'RGlyZWN0b3J5OjM=',
            'isSynchronized': true,
            'email': 'group_g@advantys.com'
          }
        }
      })
      .end(function (err, res) {
        res.should.have.status(200);
        res.should.be.json;

        res.body.should.have.property('clientMutationId');
        res.body.should.have.property('groupId');

        expect(res.body.clientMutationId).to.be.null;
        expect(res.body.groupId).to.not.be.null;

        groupId = res.body.groupId;

        done();
      });
  });

  it("should return the group created with the 'Group G' name ", function (done) {
    if (groupId === null) {
      this.skip();
    }
    chai.request(url)
      .post('/')
      .send({
        'operation': 'getGroup',
        'args': {
          'id': groupId
        }
      })
      .end(function (err, res) {
        res.should.have.status(200);
        res.should.be.json;

        utils.validateGroupProperties(res.body);

        expect(res.body).to.include({ id: groupId, name: 'Group G', description: 'Group G', directoryId: 'RGlyZWN0b3J5OjM=',
        isSynchronized: true, email: 'group_g@advantys.com' });
        done();
      });
  });

  it('should create a group with a query', function (done) {
    chai.request(url)
      .post('/')
      .send({
        'operation': 'createGroup',
        'args': {
          'input': {
            'name': 'Group H',
            'directoryId': 'RGlyZWN0b3J5OjM=',
            'query': "CITY='MONTREAL'"
          }
        }
      })
      .end(function (err, res) {
        res.should.have.status(200);
        res.should.be.json;

        res.body.should.have.property('clientMutationId');
        res.body.should.have.property('groupId');

        expect(res.body.clientMutationId).to.be.null;
        expect(res.body.groupId).to.not.be.null;

        groupId = res.body.groupId;

        done();
      });
  });

  it("should return the group created with the 'Group H' name ", function (done) {
    if (groupId === null) {
      this.skip();
    }
    chai.request(url)
      .post('/')
      .send({
        'operation': 'getGroup',
        'args': {
          'id': groupId
        }
      })
      .end(function (err, res) {
        res.should.have.status(200);
        res.should.be.json;

        utils.validateGroupProperties(res.body);

        expect(res.body).to.include({ id: groupId, name: 'Group H', directoryId: 'RGlyZWN0b3J5OjM=', query: "CITY='MONTREAL'" });

        done();
      });
  });

  it("should get an error when the group 'Group A' already exists in the directory 'WorkflowGen'", function (done) {
    chai.request(url)
      .post('/')
      .send({
        'operation': 'createGroup',
        'args': {
          'input': {
            'name': 'Group A',
            'directoryId': 'RGlyZWN0b3J5OjE='
          }
        }
      })
      .end(function (err, res) {
        res.should.have.status(500);
        res.should.be.json;

        res.body.should.have.property('error');
        expect(res.body.error).to.contain('Group already exists in the specified directory');

        done();
      });
  });

  it('should get an error when the email is not valid', function (done) {
    chai.request(url)
      .post('/')
      .send({
        'operation': 'createGroup',
        'args': {
          'input': {
            'name': 'Group I',
            'directoryId': 'RGlyZWN0b3J5OjM=',
            'email': 'group_iadvantys.com'
          }
        }
      })
      .end(function (err, res) {
        res.should.have.status(500);
        res.should.be.json;

        res.body.should.have.property('error');
        expect(res.body.error).to.contain('Invalid email');

        done();
      });
  });

  it('should get an error when query is not valid', function (done) {
    chai.request(url)
      .post('/')
      .send({
        'operation': 'createGroup',
        'args': {
          'input': {
            'name': 'Group I',
            'directoryId': 'RGlyZWN0b3J5OjM=',
            'query': 'test'
          }
        }
      })
      .end(function (err, res) {
        res.should.have.status(500);
        res.should.be.json;

        res.body.should.have.property('error');
        expect(res.body.error).to.contain('The query is not valid');

        done();
      });
  });
});
